import wx
import os
import random
import ast
import sys
if sys.version_info[0] < 3:
    import thread
import threading
import wx.aui
from PyTASAssembly import *
from PyTASRubberband import *
from SimulatorFramework import Frontier
from SimulatorFramework import SimCanvas


# IDs
PYTAS_ID_STOP          = 100
PYTAS_ID_FOR           = 101
PYTAS_ID_FF            = 102
PYTAS_ID_ZOOMIN        = 103
PYTAS_ID_ZOOMOUT       = 104
PYTAS_ID_ZOOM_MODE     = 105
PYTAS_ID_SELECT_MODE   = 106
PYTAS_ID_DELETE        = 107
PYTAS_ID_UNDO          = 108

PYTAS_SM_STEP = 0
PYTAS_SM_FULL = 1


#----------------------------------------------------------------------------

class PyTASCanvas(SimCanvas.Canvas):
    def __init__(self, parent, idIn, frame, size, app):
        self.drawUpdates = SimCanvas.SIMFW_FULLDISPLAY
        SimCanvas.Canvas.__init__(self, parent, idIn, frame, size, app)

        self.backgroundColor = wx.Colour((255, 255, 128, 255))
        self.backgroundBrush = wx.Brush(self.backgroundColor)
        self.backgroundPen = wx.Pen(self.backgroundColor)
        self.drawBitmaps = True
        self.stepMode = PYTAS_SM_STEP
        self.numTilesPerStep = 1
        self.displayMask = DISPLAY_ALL

    def CreateNewAssembly(self):
        self.sideLen = SimCanvas.CELLSIZE
        self.assembly = PyTASAssembly()
        self.UpdateSizeVars()
        #self.RedrawAssembly()
        self.CanvasSizeUpdate((0,0))
        
    def UpdateTileDisplayMask(self, bShowLabels=None, bShowGlues=None):
        savedMask = self.displayMask
        maskUpdate = DISPLAY_NONE
        if bShowLabels == True:
            self.displayMask = self.displayMask | DISPLAY_LABEL
        elif bShowLabels == False:
            self.displayMask = (~DISPLAY_LABEL) & self.displayMask 
        
        if bShowGlues == True:
            self.displayMask = self.displayMask | DISPLAY_GLUES
        elif bShowGlues == False:
            self.displayMask = (~DISPLAY_GLUES) & self.displayMask
            
        if not savedMask == self.displayMask:
            self.Refresh()

    def LoadMacrotileImage(self, mappedTypeName, tileSize):
        if not mappedTypeName in self.assembly.macrotileImages.keys():
            #MJP: Note that this section is a hard-coded hack for a particular system!!!
            mappedTypeName = mappedTypeName.rsplit('-', 1)[0]
            #if mappedTypeName.endswith("-SEED"):
            #    mappedTypeName = mappedTypeName[0:-len("-SEED")]
            imgFile = self.assembly.macrotileImageDir + "/TileImage_" + mappedTypeName + ".tif"
            try:
                img = wx.Image(imgFile, wx.BITMAP_TYPE_TIF)#.ConvertToBitmap()
                if img.HasAlpha():
                    for w in range(img.GetWidth()):
                        for h in range(img.GetHeight()):
                            if img.GetAlpha(w, h) > 5:
                                img.SetAlpha(w, h, 100)
                else:
                    print("Image '" + imgFile + "' does not have an alpha channel (for transparency)")
                #img = img.ConvertToImage()
                width = tileSize * self.assembly.macrotileWidth
                height = tileSize * self.assembly.macrotileHeight
                img = img.Scale(width, height, wx.IMAGE_QUALITY_HIGH)
                self.assembly.macrotileImages[mappedTypeName] = wx.Bitmap(img)
            except:
                SimUtils.globalApp.OutputMessage("ERROR: Failed to load macrotile image from file '" + imgFile + "'")
                self.assembly.macrotileImages[mappedTypeName] = None
        
    def DrawAtCoords(self, dc, xCoord, yCoord, loctype):
        if loctype is None:
            return
        if loctype == TileAssembly.FRONTIER_LOC or loctype == TileAssembly.FRONTIER_LOC_OFF:
            if self.assembly.showOnlyMacrotilesMode == False:
                self.DrawFrontierLocation(dc, (xCoord, yCoord), loctype)
        else:
            (xPixel,yPixel) = self.GetPixelsForCoords(xCoord, yCoord)
            # Adjust coords for correct offset of tile images
            xPixel = xPixel - self.imgOffsetHoriz
            yPixel = yPixel - self.imgOffsetVert

            #HACK to work around a bug writing to the DC multiple times on OSX
            if not "wxOSX" in wx.PlatformInfo:
                if self.assembly.showOnlyMacrotilesMode == False:
                    self.DrawFrontierLocation(dc, (xCoord, yCoord), TileAssembly.EMPTY_LOC)
            
            if (self.assembly.useMacrotileMode == True) and (not self.assembly.macrotileCenter is None) and ((xCoord % self.assembly.macrotileWidth) == self.assembly.macrotileCenter[0]) and ((yCoord % self.assembly.macrotileHeight) == self.assembly.macrotileCenter[1]):
                mappedTypeName = loctype.type.name#.split("-")[0]
                tileSize = 96
                self.LoadMacrotileImage(mappedTypeName, tileSize)
                xOffset = ((self.assembly.macrotileWidth/2)*tileSize)-tileSize/2 - 17
                yOffset = (self.assembly.macrotileHeight/2)*tileSize# + 17
                if (mappedTypeName in self.assembly.macrotileImages) and (not self.assembly.macrotileImages[mappedTypeName] is None):
                    dc.DrawBitmap(self.assembly.macrotileImages[mappedTypeName], xPixel-xOffset, yPixel-yOffset)
                else:
                    if self.assembly.showOnlyMacrotilesMode == False:
                        loctype.type.DrawAtLocation(xPixel, yPixel, dc, mask=self.displayMask)
            else:
                if self.assembly.showOnlyMacrotilesMode == False:
                    loctype.type.DrawAtLocation(xPixel, yPixel, dc, mask=self.displayMask)

    def DrawGrid(self, dc):
        dc.SetPen(self.gridPen)
        for i in range(0, self.numVert+1):
            if self.assembly.useMacrotileMode == True and self.assembly.macrotileWidth is not None:
                if i % self.assembly.macrotileWidth == 1:
                    x = i * self.sideLen
                    dc.DrawLine(x, self.height, x, self.height-self.numVert*self.sideLen) 
            else:
                x = i * self.sideLen
                dc.DrawLine(x, self.height, x, self.height-self.numVert*self.sideLen)
        for i in range(0, self.numHoriz+1):
            if self.assembly.useMacrotileMode == True and self.assembly.macrotileHeight is not None:
                if i % self.assembly.macrotileHeight == 1:
                    y = i * self.sideLen
                    dc.DrawLine(0, self.height-y, self.numHoriz*self.sideLen, self.height-y)
            else:
                y = i * self.sideLen
                dc.DrawLine(0, self.height-y, self.numHoriz*self.sideLen, self.height-y)

    def OnPaint(self, event):
        self._lock.acquire()

        dc = wx.PaintDC(self)

        bg = wx.Brush(self.backgroundColor)
        dc.SetBackground(bg)
        dc.Clear()

        # use PrepareDC to set position correctly
        self.PrepareDC(dc)
        # we need to clear the dc BEFORE calling PrepareDC
        
        clipped = dc.GetClippingBox()

        # On OS X, PrepareDC doesn't seem to account for scrolling as it does on Windows
        #if "wxOSX" in wx.PlatformInfo:
        if False: # wxpython was recently updated so that the OS specific code is not needed
                  # for Windows and Linux, and I'm guessing maybe not for OS X either
            xView,yView = self.GetViewStart()
            xDelta,yDelta = self.GetScrollPixelsPerUnit()
            x = clipped[0] + xView * xDelta
            y = clipped[1] + yView * yDelta
        else:
            x, y = self.GetViewStart()
            w, h = self.GetSize()
            topLeft = self.CoordsFromPixelPos(x, y)
            bottomRight = self.CoordsFromPixelPos(x + w, y + h)

        dc.SetUserScale(self.drawScale, self.drawScale)

        if self.drawGrid == True:
            self.DrawGrid(dc)

        #for i in range(self.assembly.minX, self.assembly.maxX+1):
            #for j in range(self.assembly.minY, self.assembly.maxY+1):
        for i in range(topLeft[0], bottomRight[0]+1):
            for j in range(bottomRight[1], topLeft[1]+1):
                loc = self.assembly.GetLoc((i,j), includeFrontier=True)
                if loc is not None:
                    if loc == TileAssembly.FRONTIER_LOC or loc == TileAssembly.FRONTIER_LOC_OFF:
                        if self.assembly.showOnlyMacrotilesMode == False:
                            self.DrawFrontierLocation(dc, (i,j), loc)
                    elif self.NameMatchesPatterns(loc.type.name) == True:
                        if self.assembly.showOnlyMacrotilesMode == False:
                            self.DrawAtCoords(dc, i, j, loc)

                        if self.assembly.useMacrotileMode == True:
                            if (not self.assembly.macrotileCenter is None) and ((i % self.assembly.macrotileWidth) == self.assembly.macrotileCenter[0]) and ((j % self.assembly.macrotileHeight) == self.assembly.macrotileCenter[1]):
                                mappedTypeName = loc.type.name#.split("-")[0]
                                tileSize = 96
                                if not mappedTypeName in self.assembly.macrotileImages:
                                    #MJP: Note that this section is a hard-coded hack for a particular system!!!
                                    mappedTypeName = mappedTypeName.rsplit('-', 1)[0]
                                    self.LoadMacrotileImage(mappedTypeName, tileSize)
                                
                                if (mappedTypeName in self.assembly.macrotileImages) and (not self.assembly.macrotileImages[mappedTypeName] is None):
                                    (xPixel,yPixel) = self.GetPixelsForCoords(i, j)
                                    # Adjust coords for correct offset of tile images
                                    xPixel = xPixel - self.imgOffsetHoriz
                                    yPixel = yPixel - self.imgOffsetVert

                                    xOffset = ((self.assembly.macrotileWidth/2)*tileSize)-tileSize/2 - 17
                                    yOffset = (self.assembly.macrotileHeight/2)*tileSize# + 17
                                    dc.DrawBitmap(self.assembly.macrotileImages[mappedTypeName], xPixel-xOffset, yPixel-yOffset)

        self._lock.release()
        self.needsRedraw = False
        self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.GetNumTiles()) + ", frontier size: " + str(len(self.assembly.frontier)))
        return

    def DoSave(self):
        #saveFileDialog = wx.FileDialog(self.window, "Save As", "", "", 
        #                               "Tile system files (*.txt)|*.txt",
        #                               wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
        #saveFileDialog.ShowModal()
        #fileName = saveFileDialog.GetPath()
        #saveFileDialog.Destroy()

        #self.saveSnapshot()

        dlg = wx.MessageDialog(self, "Sorry, this hasn't been implemented yet..")
        dlg.ShowModal()
        dlg.Destroy()
        return
    
        currFile = "output.txt"
        newFile = wx.FileSelector("Save assembly to file...", currFile, currFile, ".txt", "*.txt", wx.FD_SAVE|wx.FD_OVERWRITE_PROMPT)
        if len(newFile) > 0:
            fOut = open(newFile, "w")
            iCount = 0
            jCount = 0
            for i in range(self.numHoriz):
                for j in range(self.numVert):
                    if self.grid.IsEmpty(i, j) == False:
                        pt = self.grid.GetCenterOffsetOfType(i, j)
                        if pt.x == 0 and pt.y == 0:
                            name = self.grid.GetType(i, j)
                            outStr = name + "," + str(i) + "," + str(j) + "\n"
                            fOut.write(outStr)
            fOut.close()

    def saveSnapshot(self, image_file_name, minX, maxX, minY, maxY):
        # Find the coordinate boundaries so the canvas can be sized correctly
        min_pixels = self.GetPixelsForCoords(minX, maxY)
        max_pixels = self.GetPixelsForCoords(maxX+1, minY-1)
        
        assembly_width  = max_pixels[0] - min_pixels[0]
        assembly_height = max_pixels[1] - min_pixels[1]

        bmp = wx.Bitmap(assembly_width, assembly_height)
        memDC = wx.MemoryDC(bmp)

        bg = wx.Brush(self.backgroundColor)
        memDC.SetBackground(bg)
        memDC.Clear()

        x_offset = minX - self.assembly.minX
        y_offset = self.assembly.maxY - maxY

        for i in range(self.assembly.minX, self.assembly.maxX+1):
            for j in range(self.assembly.minY, self.assembly.maxY+1):
                loc = self.assembly.GetLoc((i,j))
                if loc is not None:
                    if self.NameMatchesPatterns(loc.type.name) == True:
                        x = i - x_offset
                        y = j + y_offset
                        self.DrawAtCoords(memDC, x, y, self.assembly.GetLoc((i,j)))

        image_type = image_file_name[-3:]
        if image_type == 'bmp':
            image_type = wx.BITMAP_TYPE_BMP
        elif image_type == 'png':
            image_type = wx.BITMAP_TYPE_PNG
        elif image_type == 'jpg':
            image_type = wx.BITMAP_TYPE_JPEG
        #elif image_type == 'gif':
        #    image_type = wx.BITMAP_TYPE_GIF
        else:
            print(f'ERROR: Invalid image type "{image_type}"')
            return
        
        # Select the Bitmap out of the memory DC by selecting a new uninitialized Bitmap
        memDC.SelectObject(wx.NullBitmap)
        bmp.SaveFile(image_file_name, image_type)

    def ShowHelp(self):
        info = wx.adv.AboutDialogInfo()
        info.SetName("PyTAS (Python-based Tile Assembly Simulator)")
        info.SetVersion(self.app.GetVersionString())
        helpStr = "For general information about the abstract Tile Assembly Model (aTAM)\n"
        helpStr = helpStr + "or to learn more about ISU TAS, a C++ based aTAM simulator upon\n"
        helpStr = helpStr + "which PyTAS is based, please see http://www.self-assembly.net\n"
        helpStr = helpStr + "\nList of keyboard commands:\n\n"
        helpStr = helpStr + "  L : Load a tile assembly system\n"
        helpStr = helpStr + "  S : Save current tile assembly system\n"
        helpStr = helpStr + "  R : Reset current tile assembly system to seed\n"
        helpStr = helpStr + "  F : Simulate one step\n"
        helpStr = helpStr + "  G : Begin fast forward simulation\n"
        helpStr = helpStr + "  Q : Quit fast forward simulation\n"
        #helpStr = helpStr + "  p : print list of frontier locations\n"
        helpStr = helpStr + "  T : Toggle display of grid lines\n"
        helpStr = helpStr + "  D : Toggle display updates of tiles to assembly\n"
        helpStr = helpStr + "  C : Center screen on desired coordinates"
        helpStr = helpStr + " +/-: Zoom in/out\n"
        helpStr = helpStr + "<left mouse click>:  Zoom in on clicked location\n"
        helpStr = helpStr + "<right mouse click>:  Zoom out of clicked location\n"
        helpStr = helpStr + "<F2> : Refresh display"
        helpStr = helpStr + "  H : show this help dialog"
        info.SetDescription(helpStr)
        wx.adv.AboutBox(info)
