from PyTASTileType import *
import copy
import os
import sys
import traceback

class TDS:
	name = ""
	tileList = []
	tileDictionary = {}

	def __init__(self, name, tileList = [], tileDictionary = {}):
		if ".tds" not in name:
			name += ".tds"
		self.name = name
		if tileList == []:
			self.tileList = list()
		else:
			self.tileList = tileList
		if tileDictionary == {}:
			self.tileDictionary = dict()
		else:
			self.tileDictionary = tileDictionary

	def Write(self):
		if not os.path.exists("./TDPTDS"):
			os.makedirs("./TDPTDS")
		if ".tds" not in self.name:
			self.name += ".tds"
		with open("./TDPTDS/" + self.name, "w") as outFile:
			for tile in self.tileList:
				outFile.write(tile.Output())
				outFile.write("\n")
		print "Wrote " + str(len(self.tileList)) + " tiles to " + self.name 

	def Merge(self, otherTDS):
		self.tileList += otherTDS.tileList
		self.tileDictionary.update(otherTDS.tileDictionary)

	def AddTiles(self, tileList):
		self.tileList += tileList
		for tile in tileList:
			self.tileDictionary[tile.name] = tile

	def AddTile(self, tile):
		debug = True
		#debug = False
		if self.tileDictionary.get(tile.name) != None:
			if(debug):
				print "Duplicate Tile Error: " + tile.name
				raise TypeError("Duplicate Tile")
				exc_type, exc_value, exc_traceback = sys.exc_info()
				print "*** print_exception:"
				traceback.print_exception(exc_type, exc_value, exc_traceback, limit=5, file=sys.stdout)
				exit(-1)
			print "A tile with name \"" + tile.name + "\" already exists in " + self.name + "\n"
			i = 1
			while self.tileDictionary.get(tile.name + "(" + str(i) + ")") != None:
				i += 1
			print "Name: \"" + tile.name + "\" was changed to \"" + tile.name + "(" + str(i) + ")\"\n"
			tile.name = tile.name + "(" + str(i) + ")" 
		self.tileDictionary[tile.name] = copy.deepcopy(tile)
		self.tileList.append(copy.deepcopy(tile))	

def parseTDS(tdsPath):
	tdsFileName = tdsPath.split("/")[-1]
	tds = TDS(tdsFileName)


	if ".tds" not in tdsFileName:
		print "Invalid File Name: Lacking .tds suffix"
		return None

	l = []
	dictionary = {}
	with open(tdsPath, "r") as inFile:
		content = inFile.readlines()
		for line in content:

			#Pre-Processing
			line = line.strip("\n")
			lineSplit = line.split(" ")
			lineInstruction = lineSplit[0]
			if len(lineSplit) > 1:
				lineData = lineSplit[1]

			#Parse to Tile
			if "TILENAME" in lineInstruction:
				t = TileType(True)
				t.name = lineData
				continue
			elif "LABEL" in lineInstruction and lineInstruction not in [dirName + "LABEL" for dirName in DirNameArray]:
				t.label = lineData
				continue
			elif "TILECOLOR" in lineInstruction:
				t.color = " ".join(lineSplit[1:])
				continue
			elif "CREATE" in lineInstruction:
				l.append(t)
				dictionary[t.name] = t
				continue
			else:
				for dirNum, dirName in enumerate(DirNameArray):
					if dirName.upper() in line:
						if "BIND" in line:
							t.sides[dirNum].strength = int(lineData)
							break
						if "LABEL" in line:
							t.sides[dirNum].label = lineData
							break
		tds.tileList = l
		tds.tileDictionary = dictionary
	return tds

def parseTDSList(tdsFileList):
	tdsList = []
	for tdsFileName in tdsFileList:
		tds = TDS(tdsFileName)
		tds.parseTilesFromTDS(tdsFileName)
		tdsList.append(tds)
	return tdsList

