from PyTASTileType import *
import math
import sys
import getopt


def DisplayHelp():
    print("\nThis script takes as input a number, n, for the side lengths of a square and possibly the names")
    print("of output files, and outputs files for an aTAM system which generates an n x n square.")
    print("Note: The generated system will be correct for values of n > 3.")
    print("\nUsage:")
    print("-h, --help\t\t\t\tDisplays this help message.")
    print("-n <int>, --n=<int>\t\t\tSide lengths of the square to be built.")
    print("-s <tdp file>, --seed=<tdp file>\t[Optional] TDP file to output. If not specified,")
    print("\t\t\t\t\t`<n>x<n>_square.tdp' will be used for given value n.")
    print("-t <tds file>, --tileset=<tds file>\t[Optional] TDS file to output. If not specified,")
    print("\t\t\t\t\t`<n>x<n>_square.tds' will be used for given value n.")
    print("-d <dir>, --dir=<dir>\t\t\t[Optional] Directory for output files. If not specified,")
    print("\t\t\t\t\tthe current directory will be used.")
    print("")
    exit()


def GetSeed(bit_string):
    seed_tiles = []
    loc = 0
    start_msb = True
    for bit in bit_string[::-1]:
        bit = int(bit)
        #print(f'{bit}')
        t = PyTASTileType()
        t.name = f'Seed{loc}'
        if loc > 0:
            t.sides[E].strength = 2
            t.sides[E].label = f'S{loc}'
        if loc < (len(bit_string) - 1):
            t.sides[W].strength = 2
            t.sides[W].label = f'S{loc+1}'
        if loc == 0:
            if bit == 1:
                t.sides[N].strength = 2
                start_msb = False
            else:
                t.sides[N].strength = 1
            t.sides[N].label = f'L{bit}'
            t.label = f'L{bit}'
            t.sides[E].label = 'fill*'
            t.sides[E].strength = 2
            t.sides[S].label = '*fill'
            t.sides[S].strength = 2
        elif (loc == (len(bit_string) - 1)):
            t.sides[N].label = f'M{bit}'
            t.label = f'M{bit}'
            if start_msb:
                t.sides[N].strength = 2
            else:
                t.sides[N].strength = 1
            t.sides[S].label = 'fill'
            t.sides[S].strength = 1
        else:
            t.sides[N].label = f'{bit}'
            t.sides[N].strength = 1
            t.label = f'{bit}'
            t.sides[S].label = 'fill'
            t.sides[S].strength = 1
        loc += 1
        seed_tiles.append(t)

    for st in seed_tiles:
        st.color = 'Green'
    return seed_tiles

def GetCounterTileTypes():
    tile_types = []

    #################3
    # LSB tiles

    # bit = 0
    t = PyTASTileType()
    t.name = 'L0'
    t.label = 'L0'
    t.sides[S].label = 'L1'
    t.sides[S].strength = 2
    t.sides[W].label = '<1'
    t.sides[W].strength = 1
    t.sides[N].label = 'L0'
    t.sides[N].strength = 1
    t.sides[E].label = 'fill'
    t.sides[E].strength = 1
    tile_types.append(t)

    # bit = 1
    t = PyTASTileType()
    t.name = 'L1'
    t.label = 'L1'
    t.sides[S].label = 'L0'
    t.sides[S].strength = 1
    t.sides[W].label = '0>'
    t.sides[W].strength = 1
    t.sides[N].label = 'L1'
    t.sides[N].strength = 2
    t.sides[E].label = 'fill'
    t.sides[E].strength = 1
    tile_types.append(t)


    ##################
    # middle tiles

    ###############
    # right to left
    
    # 00
    t = PyTASTileType()
    t.name = '<00'
    t.label = '0'
    t.sides[E].label = '<0'
    t.sides[E].strength = 1
    t.sides[S].label = '0'
    t.sides[S].strength = 1
    t.sides[W].label = '<0'
    t.sides[W].strength = 1
    t.sides[N].label = '0'
    t.sides[N].strength = 1
    tile_types.append(t)

    # 01
    t = PyTASTileType()
    t.name = '<01'
    t.label = '1'
    t.sides[E].label = '<1'
    t.sides[E].strength = 1
    t.sides[S].label = '0'
    t.sides[S].strength = 1
    t.sides[W].label = '<0'
    t.sides[W].strength = 1
    t.sides[N].label = '1'
    t.sides[N].strength = 1
    tile_types.append(t)

    # 10
    t = PyTASTileType()
    t.name = '<10'
    t.label = '1'
    t.sides[E].label = '<0'
    t.sides[E].strength = 1
    t.sides[S].label = '1'
    t.sides[S].strength = 1
    t.sides[W].label = '<0'
    t.sides[W].strength = 1
    t.sides[N].label = '1'
    t.sides[N].strength = 1
    tile_types.append(t)

    # 11
    t = PyTASTileType()
    t.name = '<11'
    t.label = '0'
    t.sides[E].label = '<1'
    t.sides[E].strength = 1
    t.sides[S].label = '1'
    t.sides[S].strength = 1
    t.sides[W].label = '<1'
    t.sides[W].strength = 1
    t.sides[N].label = '0'
    t.sides[N].strength = 1
    tile_types.append(t)

    ###############
    # left to right

    # 00
    t = PyTASTileType()
    t.name = '00>'
    t.label = '0'
    t.sides[E].label = '0>'
    t.sides[E].strength = 1
    t.sides[S].label = '0'
    t.sides[S].strength = 1
    t.sides[W].label = '0>'
    t.sides[W].strength = 1
    t.sides[N].label = '0'
    t.sides[N].strength = 1
    tile_types.append(t)

    # 10
    t = PyTASTileType()
    t.name = '10>'
    t.label = '1'
    t.sides[E].label = '0>'
    t.sides[E].strength = 1
    t.sides[S].label = '1'
    t.sides[S].strength = 1
    t.sides[W].label = '0>'
    t.sides[W].strength = 1
    t.sides[N].label = '1'
    t.sides[N].strength = 1
    tile_types.append(t)

    #################3
    # MSB tiles

    # row start, bit = 0
    t = PyTASTileType()
    t.name = 'M0>'
    t.label = 'M0'
    t.sides[S].label = 'M0'
    t.sides[S].strength = 2
    t.sides[E].label = '0>'
    t.sides[E].strength = 1
    t.sides[N].label = 'M0'
    t.sides[N].strength = 1
    tile_types.append(t)

    # row start, bit = 1
    t = PyTASTileType()
    t.name = 'M1>'
    t.label = 'M1'
    t.sides[S].label = 'M1'
    t.sides[S].strength = 2
    t.sides[E].label = '0>'
    t.sides[E].strength = 1
    t.sides[N].label = 'M1'
    t.sides[N].strength = 1
    tile_types.append(t)

    # row end, bit = 0
    t = PyTASTileType()
    t.name = '<M0'
    t.label = 'M0'
    t.sides[S].label = 'M0'
    t.sides[S].strength = 1
    t.sides[E].label = '<0'
    t.sides[E].strength = 1
    t.sides[N].label = 'M0'
    t.sides[N].strength = 2
    tile_types.append(t)

    # row end, bit = 1
    t = PyTASTileType()
    t.name = '<M1'
    t.label = 'M1'
    t.sides[S].label = 'M1'
    t.sides[S].strength = 1
    t.sides[E].label = '<0'
    t.sides[E].strength = 1
    t.sides[N].label = 'M1'
    t.sides[N].strength = 2
    tile_types.append(t)

    # row end, bit file to 1
    t = PyTASTileType()
    t.name = '<M1c'
    t.label = 'M1'
    t.sides[S].label = 'M0'
    t.sides[S].strength = 1
    t.sides[E].label = '<1'
    t.sides[E].strength = 1
    t.sides[N].label = 'M1'
    t.sides[N].strength = 2
    tile_types.append(t)


    # final counter tile
    t = PyTASTileType()
    t.name = 'End'
    t.label = 'End'
    t.sides[S].label = 'M1'
    t.sides[S].strength = 1
    t.sides[E].label = '<1'
    t.sides[E].strength = 1
    tile_types.append(t)

    for tt in tile_types:
        tt.color = 'Yellow'
    return tile_types


def GetFillerTileTypes():
    tile_types = []

    t = PyTASTileType()
    t.name = 'fill'
    t.label = 'fill'
    t.sides[S].label = 'fill'
    t.sides[S].strength = 1
    t.sides[W].label = 'fill'
    t.sides[W].strength = 1
    t.sides[N].label = 'fill'
    t.sides[N].strength = 1
    t.sides[E].label = 'fill'
    t.sides[E].strength = 1
    tile_types.append(t)

    t = PyTASTileType()
    t.name = 'fill2'
    t.label = 'fill'
    t.sides[S].label = 'fill>'
    t.sides[S].strength = 1
    t.sides[W].label = 'fill>'
    t.sides[W].strength = 1
    t.sides[N].label = 'fill>'
    t.sides[N].strength = 1
    t.sides[E].label = 'fill>'
    t.sides[E].strength = 1
    tile_types.append(t)

    t = PyTASTileType()
    t.name = 'fill>'
    t.label = 'fill>'
    t.sides[S].label = 'fill>'
    t.sides[S].strength = 1
    t.sides[W].label = 'fill*'
    t.sides[W].strength = 2
    t.sides[N].label = 'fill*'
    t.sides[N].strength = 1
    t.sides[E].label = 'fill>'
    t.sides[E].strength = 1
    tile_types.append(t)

    t = PyTASTileType()
    t.name = 'fill*'
    t.label = 'fill*'
    t.sides[S].label = 'fill*'
    t.sides[S].strength = 1
    t.sides[W].label = 'fill'
    t.sides[W].strength = 1
    t.sides[N].label = 'fill'
    t.sides[N].strength = 1
    t.sides[E].label = 'fill*'
    t.sides[E].strength = 2
    tile_types.append(t)

    # fill under the seed
    t = PyTASTileType()
    t.name = '*fill'
    t.label = '*fill'
    t.sides[S].label = 'fill>'
    t.sides[S].strength = 1
    t.sides[W].label = '*fill'
    t.sides[W].strength = 1
    t.sides[N].label = '*fill'
    t.sides[N].strength = 2
    t.sides[E].label = 'fill>'
    t.sides[E].strength = 1
    tile_types.append(t)

    t = PyTASTileType()
    t.name = '<fill'
    t.label = '*fill'
    t.sides[S].label = '*fill'
    t.sides[S].strength = 2
    t.sides[W].label = 'fill'
    t.sides[W].strength = 1
    t.sides[N].label = 'fill'
    t.sides[N].strength = 1
    t.sides[E].label = '*fill'
    t.sides[E].strength = 1
    tile_types.append(t)

    for tt in tile_types:
        tt.color = 'Grey'
    return tile_types


if __name__ == "__main__":
    try:
        # starts at the second element of argv since the first one is the script name
        # extraparams are extra arguments passed after all option/keywords are assigned
        # opts is a list containing the pair "option"/"value"
        opts, extraparams = getopt.getopt(sys.argv[1:], "hn:s:t:d:",
                            ["help", "n=", "seed=", "tileset=", "dir="])
        #print 'Opts:',opts
        #print 'Extra parameters:',extraparams
    except:
        print("")
        print("Unrecognized option:" + str(sys.argv[1:]))
        DisplayHelp()

    if len(opts) == 0:
        DisplayHelp()

    side_len = None
    tdp_file = None
    tds_file = None
    directory = '.'

    for o,p in opts:
        if o in ['-h', '--help']:
            DisplayHelp()
        elif o in ['-n', '--n']:
            side_len = int(p)
        elif o in ['-s', '--seed']:
            tdp_file = p
        elif o in ['-t', '--tileset']:
            tds_file = p
        elif o in ['-d', '--dir']:
            directory = p
        else:
            DisplayHelp()

    if side_len is None:
        print("\nNo side length specified!")
        DisplayHelp()
    if tdp_file is None:
        tdp_file = f'{side_len}x{side_len}_square.tdp'
    if tds_file is None:
        tds_file = f'{side_len}x{side_len}_square.tds'


    seed_width = math.ceil(math.log(side_len,2))
    seed_val = int(math.pow(2, seed_width) - (side_len - seed_width - 1))
    seed_num  = bin(seed_val)[2:].zfill(seed_width)
    print(f'For desired side length {side_len}: counter width = {seed_width}, seed row value = {seed_val}, seed row binary value = {seed_num}')


    seed_tiles = GetSeed(seed_num)
    print(f'Number of unique seed tile types: {len(seed_tiles)}')
    counter_tiles = GetCounterTileTypes()
    filler_tiles = GetFillerTileTypes()
    print(f'Number of generic tile types: {len(counter_tiles+filler_tiles)} ({len(counter_tiles)} counter tiles and {len(filler_tiles)} filler tiles)')
    tile_set = seed_tiles + counter_tiles + filler_tiles

    with open(f'{directory}/{tdp_file}', 'w') as f:
        f.write(f'{tds_file}\n')
        f.write(f'{seed_tiles[0].name} {seed_width-1} {seed_width}\n')

    with open(f'{directory}/{tds_file}', 'w') as f:
        for tile in tile_set:
            f.write(f'{tile.Output()}\n')
    print('Tile system output complete.')


